package executor;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import objects.FrameJointCoordinates;
import objects.MotionJointCoordinates;
import objects.FrameBoneRotations;
import objects.MotionBoneRotations;
import objects.Joint;
import objects.Trajectory;
import objects.Bone;
import objects.Feature;
import objects.Motion;
import objects.Skeleton;
import objects.Trajectory3d;
import objects.Triple;
import methods.Method;
import methods.MethodAhmed;
import methods.MethodAli;
import methods.MethodAndersson;
import methods.MethodBall;
import methods.MethodDikovski;
import methods.MethodGavrilova;
import methods.MethodJiang;
import methods.MethodKrzeszowski;
import methods.MethodKumar;
import methods.MethodKwolek;
import methods.MethodPreis;
import methods.MethodSedmidubsky;
import methods.MethodSinha;
import methods.Method_MMC;
import methods.Method_PCALDA;
import methods.Method_Random;
import methods.Method_Raw;
import objects.ClassifiableList;
import objects.Parameter;
import objects.Template;

public class Executor {

    /*
     File locations
     */
    private static final File prototypicalFileASF = new File("skeleton.asf");
    private static final File prototypicalFileAMC = new File("gaitcycle.amc");
    private static final File customQueryFileAMC = new File("query.amc");
    private static final File customGalleryDirectory = new File("gallery/");
    private static final File customClassifier = new File("classifiers/_MMC-302.0.classifier");

    /*
     Select distance threshold
     Each value refers to a particular sub-database
     */
    private static final double distanceThreshold = 302.0; // 56.3(2,35), 59.4(4,67), 63.3(8,130), 73.7(16,302), 173.3(32,2047), 302.0(54,3843), 495.3(64,5923)

    /*
     Select actions
     */
    private static final boolean extractDatabase = false;
    private static final boolean learnClassifiers = false;
    private static final boolean performClassifications = false;
    private static final boolean evaluateMethods = true;

    /*
     Select evaluation set-ups at evaluateMethods()
     */
    private static final boolean evaluateHomogeneous = true;
    private static final boolean evaluateHeterogeneous = false;

    /*
     All implemented MoCap-based gait recognition methods
     */
    private static final Method[] methods = new Method[]{
        new MethodAhmed(),//0
        new MethodAli(),//1
        new MethodAndersson(),//2
        new MethodBall(),//3
        new MethodDikovski(),//4
        new MethodGavrilova(),//5
        new MethodJiang(),//6
        new MethodKrzeszowski(),//7
        new MethodKumar(),//8
        new MethodKwolek(),//9
        new MethodPreis(),//10
        new MethodSedmidubsky(),//11
        new MethodSinha(),//12
        new Method_MMC(true),//13
        new Method_MMC(false),//14
        new Method_PCALDA(true),//15
        new Method_PCALDA(false),//16
        new Method_Random(),//17
        new Method_Raw(true),//18
        new Method_Raw(false),//19
    };

    /*
     Main method
     Executes the actions selected above
     */
    public static void main(String[] args) throws IOException, ClassNotFoundException {
        if (extractDatabase) {
            extractDatabase();
        }
        if (learnClassifiers) {
            learnClassifiers();
        }
        if (performClassifications) {
            performClassifications();
        }
        if (evaluateMethods) {
            evaluateMethods();
        }
    }

    /*
     Normalize the whole original CMU database in amcOriginal/ directory with respect to person’s position and walk direction and saves it to amc/ directory
     Select gait cycles as sub-motions with similarity under selected distanceThreshold to prototypicalFileAMC and save them to separate AMC files {subject}_{id}[{from}-{to}].amc in amc{distanceThreshold}/ directory
     Only subjects of at least 10 gait cycles are kept.
     */
    private static void extractDatabase() throws IOException {
        normalizeDatabase();
        extractGaitCycles();
    }

    /*
     Extracts classifiers for all methods learned on the sub-database determined by distanceThreshold
     Classifiers are saved as separate files {methodName}-{distanceThreshold}.classifier in classifiers/ directory
     */
    private static void learnClassifiers() throws IOException {
        new File("classifiers/").mkdirs();
        for (Method method : methods) {
            method.learnClassifier(method.loadMotions(new File("amc" + distanceThreshold + "/")));
            method.saveClassifier(new File("classifiers/" + method.getName() + "-" + distanceThreshold + ".classifier"));
        }
    }

    /*
     Performs classifications of customQueryFileAMC over customGalleryDirectory for all methods
     Instead of learning classifiers methods load their extracted classifiers {distanceThreshold}-{methodName}.classifier in classifiers/ directory that are learned on the sub-database determined by distanceThreshold
     */
    private static void performClassifications() throws IOException, ClassNotFoundException {
        for (Method method : methods) {
            method.loadClassifier(customClassifier);
            Template templateQuery = method.extractTemplate(method.loadMotion(customQueryFileAMC));
            List<Template> templatesGallery = method.extractTemplates(method.loadMotions(customGalleryDirectory));
            System.out.print(method.getName() + ": query subject classified as " + method.classify(templateQuery, templatesGallery) + "\r\n");
        }
    }

    /*
     Performs evaluations for all methods
     */
    private static void evaluateMethods() throws IOException {
        long beginningTotal = System.currentTimeMillis();
        for (Method method : methods) {
            System.out.print(method.getName() + "," + distanceThreshold + "\r\n");
            List<Motion> motions = method.loadMotions(new File("amc" + distanceThreshold + "/"));
            int numberOfClasses = new ClassifiableList(motions).splitClasses().size();
            method.setNFoldsEvaluation(10); // number of folds in evaluation data
            method.setFineness(30); // fineness of FAR/FRR, GAR/IAR, RCL/PCN
            if (evaluateHomogeneous) {
                int nClasses; // number of learning and evaluation classes
                int nFoldsLearning; // number of learning folds
                int attempts; // number of evaluation attempts
                long beginning; // time of beginning

                // default
                nClasses = numberOfClasses;
                nFoldsLearning = 3;
                attempts = 1;
                beginning = System.currentTimeMillis();
                method.evaluateHomogeneous(motions, nClasses, nFoldsLearning, attempts);
                System.out.print("timeHomogeneous=" + printTime(System.currentTimeMillis() - beginning) + "\r\n\r\n");

                // S+SSPR Figure 3
//                for (int i = 2; i <= numberOfClasses / 2; i++) {
//                    nClasses = i;
//                    nFoldsLearning = 3;
//                    attempts = 1;
//                    System.out.print("(" + nClasses + "," + nClasses +")\r\n");
//                    beginning = System.currentTimeMillis();
//                    method.evaluateHomogeneous(motions, nClasses, nFoldsLearning, attempts);
//                    System.out.print("timeHomogeneous=" + printTime(System.currentTimeMillis() - beginning) + "\r\n\r\n");
//                }
            }
            if (evaluateHeterogeneous) {
                int nClassesLearning; // number of learning classes
                int nClassesEvaluation; // number of evaluation classes
                int attempts; // number of evaluation attempts
                long beginning; // time of beginning

                // default
                nClassesLearning = 2; // number of learning classes
                nClassesEvaluation = numberOfClasses - 2;
                attempts = 3;
                beginning = System.currentTimeMillis();
                method.evaluateHeterogeneous(motions, nClassesLearning, nClassesEvaluation, attempts);
                System.out.print("timeHeterogeneous=" + printTime(System.currentTimeMillis() - beginning) + "\r\n\r\n");
                
                // S+SSPR Figure 3
//                for (int i = 2; i <= numberOfClasses / 2; i++) {
//                    nClassesLearning = i;
//                    nClassesEvaluation = i;
//                    attempts = 3;
//                    beginning = System.currentTimeMillis();
//                    System.out.print("(" + nClassesLearning + ";" + nClassesEvaluation +")\r\n");
//                    method.evaluateHeterogeneous(motions, nClassesLearning, nClassesEvaluation, attempts);
//                    System.out.print("timeHeterogeneous=" + printTime(System.currentTimeMillis() - beginning) + "\r\n\r\n");
//                }
                // S+SSPR Figure 4a
//                for (int i = 2; i <= numberOfClasses / 2; i++) {
//                    nClassesLearning = i;
//                    nClassesEvaluation = numberOfClasses / 2;
//                    attempts = 3;
//                    beginning = System.currentTimeMillis();
//                    System.out.print("(" + nClassesLearning + ";" + nClassesEvaluation +")\r\n");
//                    method.evaluateHeterogeneous(motions, nClassesLearning, nClassesEvaluation, attempts);
//                    System.out.print("timeHeterogeneous=" + printTime(System.currentTimeMillis() - beginning) + "\r\n\r\n");
//                }
                // S+SSPR Figure 4b
//                for (int i = 2; i <= numberOfClasses - 2; i++) {
//                    nClassesLearning = i;
//                    nClassesEvaluation = numberOfClasses - i;
//                    attempts = 3;
//                    beginning = System.currentTimeMillis();
//                    System.out.print("(" + nClassesLearning + ";" + nClassesEvaluation + ")\r\n");
//                    method.evaluateHeterogeneous(motions, nClassesLearning, nClassesEvaluation, attempts);
//                    System.out.print("timeHeterogeneous=" + printTime(System.currentTimeMillis() - beginning) + "\r\n\r\n");
//                }
            }
        }
        System.out.print("timeTotal=" + printTime(System.currentTimeMillis() - beginningTotal));
    }

    private static void normalizeDatabase() throws IOException {
        new File("amc/").mkdirs();
        for (File fileAMC : new File("amcOriginal/").listFiles()) {
            BufferedReader brAMC = new BufferedReader(new FileReader(fileAMC));
            BufferedWriter wrAMC = new BufferedWriter(new FileWriter("amc/" + fileAMC.getName()));
            String lineAMC = brAMC.readLine();
            while (lineAMC != null) {
                if (lineAMC.split(" ")[0].equals("root")) {
                    wrAMC.write("root 0 0 0 0 0 0\r\n");
                } else {
                    wrAMC.write(lineAMC + "\r\n");
                }
                lineAMC = brAMC.readLine();
            }
            wrAMC.flush();
            wrAMC.close();
        }
    }

    private static void extractGaitCycles() throws IOException {
        MotionJointCoordinates motionProto = convertASFAMCtoMJC(prototypicalFileASF, prototypicalFileAMC);
        Feature featureProto = motionProto.extractFeetDistanceSideSensitiveFeature();
        int lengthProto = featureProto.getLength();
        double sigma = 2;
        int minLength = (int) Math.round(lengthProto / sigma);
        int maxLength = (int) Math.round(lengthProto * sigma);
        List<Motion> gaitCycles = new ArrayList();
        for (File fileAMC : new File("amc/").listFiles()) {
            System.out.print("Processing " + fileAMC.getName() + "\r\n");
            MotionJointCoordinates motion = convertASFAMCtoMJC(prototypicalFileASF, fileAMC);
            Feature feature = motion.extractFeetDistanceSideSensitiveFeature();
            int length = feature.getLength();
            boolean selected = true;
            while (selected) {
                selected = false;
                double minDistance = Double.MAX_VALUE;
                int minFrom = 0;
                int minTo = 0;
                for (int from = 0; from < length; from++) {
                    for (int to = from + minLength; to < Math.min(from + maxLength, length); to++) {
                        Feature subFeature = feature.getSubFeature(from, to);
                        subFeature.adjust(lengthProto);
                        double[] subFeatureParameterValues = subFeature.getParameterValues();
                        double[] featureProtoParameterValues = featureProto.getParameterValues();
                        double distance = 0;
                        for (int l = 0; l < lengthProto; l++) {
                            distance += Math.abs(subFeatureParameterValues[l] - featureProtoParameterValues[l]);
                        }
                        if (minDistance > distance) {
                            minDistance = distance;
                            minFrom = from;
                            minTo = to;
                        }
                    }
                }
                if (minDistance < distanceThreshold) {
                    gaitCycles.add(motion.getSubMotion(minFrom, minTo));
                    selected = true;
                    for (int l = minFrom; l < minTo; l++) {
                        feature.setParameter(l, new Parameter("Erased", new Random().nextDouble() * Double.MAX_VALUE));
                    }
                }
            }
        }
        List<List<Motion>> gaitCyclesByClass = new ClassifiableList(gaitCycles).splitClasses();
        new File("amc" + distanceThreshold + "/").mkdirs();
        for (List<Motion> gaitCyclesOfClass : gaitCyclesByClass) {
            if (gaitCyclesOfClass.size() >= 10) {
                for (Motion gaitCycle : gaitCyclesOfClass) {
                    String stringSubjectId = String.format("%02d", gaitCycle.getSubject()) + "_" + gaitCycle.getId();
                    BufferedReader brAMC = new BufferedReader(new FileReader("amc/" + stringSubjectId.split("\\[")[0] + ".amc"));
                    BufferedWriter wrAMC = new BufferedWriter(new FileWriter("amc" + distanceThreshold + "/" + stringSubjectId + ".amc"));
                    String from = stringSubjectId.split("-")[0].split("\\[")[1];
                    String to = stringSubjectId.split("-")[1].split("\\]")[0];
                    String lineAMC = brAMC.readLine();
                    while (lineAMC != null && !lineAMC.equals("1")) {
                        wrAMC.write(lineAMC + "\r\n");
                        lineAMC = brAMC.readLine();
                    }
                    while (lineAMC != null && !lineAMC.equals(String.valueOf(from))) {
                        lineAMC = brAMC.readLine();
                    }
                    int f = 0;
                    while (lineAMC != null && !lineAMC.equals(to)) {
                        if (lineAMC.split(" ").length == 1) {
                            wrAMC.write(++f + "\r\n");
                        } else {
                            wrAMC.write(lineAMC + "\r\n");
                        }
                        lineAMC = brAMC.readLine();
                    }
                    wrAMC.flush();
                    wrAMC.close();
                    brAMC.close();
                }
            }
        }
    }

    public static List<Motion> loadMotionsJointCoordinates(File directoryAMC) throws IOException {
        List<Motion> motionsMJC = new ArrayList();
        for (File fileAMC : directoryAMC.listFiles()) {
            motionsMJC.add(convertASFAMCtoMJC(prototypicalFileASF, fileAMC));
        }
        return motionsMJC;
    }

    public static List<Motion> loadMotionsBoneRotations(File directoryAMC) throws IOException {
        List<Motion> motionsMBR = new ArrayList();
        for (File fileAMC : directoryAMC.listFiles()) {
            motionsMBR.add(convertAMCtoMBR(fileAMC));
        }
        return motionsMBR;
    }

    public static Motion loadMotionJointCoordinates(File fileAMC) throws IOException {
        return convertASFAMCtoMJC(prototypicalFileASF, fileAMC);
    }

    public static Motion loadMotionBoneRotations(File fileAMC) throws IOException {
        return convertAMCtoMBR(fileAMC);
    }

    private static MotionJointCoordinates convertASFAMCtoMJC(File fileASF, File fileAMC) throws IOException {
        Skeleton skeleton = new Skeleton();
        skeleton.readASF(fileASF.getPath());
        skeleton.readAMC(fileAMC.getPath());
        List<Trajectory> trajectories = new ArrayList();
        for (Bone bone : skeleton.getBones()) {
            Trajectory3d trajectory3d = skeleton.getTrajectory3d(bone.getName());
            List<Triple> positions3d = trajectory3d.getPositions();
            List<Triple> positions = new ArrayList();
            for (int l = 0; l < trajectory3d.getPositions().size(); l++) {
                positions.add(positions3d.get(l));
            }
            trajectories.add(new Trajectory(trajectory3d.getName(), positions));
        }
        List<FrameJointCoordinates> frames = new ArrayList();
        for (int l = 0; l < skeleton.getNumberOfFrames() - 1; l++) {
            List<Joint> joints = new ArrayList();
            joints.add(new Joint("root", skeleton.getRoot().getPositions().get(l)));
            for (Bone bone : skeleton.getBones()) {
                Trajectory trajectory = trajectories.get(0);
                for (Trajectory traj : trajectories) {
                    if (traj.getName().equals(bone.getName())) {
                        trajectory = traj;
                    }
                }
                joints.add(new Joint(bone.getName(), trajectory.getPosition(l)));
            }
            frames.add(new FrameJointCoordinates(joints));
        }
        String[] split = fileAMC.getName().split("\\.")[0].split("_");
        int subject = 0;
        String id = "";
        if (split.length == 2) {
            subject = Integer.parseInt(split[0]);
            id = split[1];
        }
        return new MotionJointCoordinates(subject, id, frames);
    }

    private static MotionBoneRotations convertAMCtoMBR(File fileAMC) throws IOException {
        String[] excludedBones = new String[]{"root"};
        List<FrameBoneRotations> frames = new ArrayList();
        BufferedReader brAMC = new BufferedReader(new FileReader(fileAMC));
        String lineAMC = brAMC.readLine();
        while (!lineAMC.equals("1")) {
            lineAMC = brAMC.readLine();
        }
        lineAMC = brAMC.readLine();
        while (lineAMC != null) {
            List<Double> boneRotationsList = new ArrayList();
            while (lineAMC != null && lineAMC.split(" ").length != 1) {
                String[] split = lineAMC.split(" ");
                boolean notExcludedBone = true;
                for (String excludedBone : excludedBones) {
                    if (split[0].equals(excludedBone)) {
                        notExcludedBone = false;
                    }
                }
                if (notExcludedBone) {
                    for (int i = 1; i < split.length; i++) {
                        boneRotationsList.add(Double.parseDouble(split[i]));
                    }
                }
                lineAMC = brAMC.readLine();
            }
            double[] boneRotations = new double[boneRotationsList.size()];
            for (int i = 0; i < boneRotationsList.size(); i++) {
                boneRotations[i] = boneRotationsList.get(i);
            }
            frames.add(new FrameBoneRotations(boneRotations));
            lineAMC = brAMC.readLine();
        }
        String[] split = fileAMC.getName().split("\\.")[0].split("_");
        int subject = 0;
        String id = "";
        if (split.length == 2) {
            subject = Integer.parseInt(split[0]);
            id = split[1];
        }
        return new MotionBoneRotations(subject, id, frames);
    }

    public static String printTime(double time) {
        String unit = "ms";
        if (time >= 1000) {
            time /= 1000;
            unit = "s";
            if (time >= 60) {
                time /= 60;
                unit = "m";
                if (time >= 60) {
                    time /= 60;
                    unit = "h";
                    if (time >= 24) {
                        time /= 24;
                        unit = "d";
                    }
                }
            }
        }
        return String.valueOf((double) Math.round(time * 10) / 10) + unit;
    }
}
