package methods;

import Jama.Matrix;
import algorithms.Classifier;
import algorithms.RetrieverRangeQuery;
import executor.Executor;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import java.util.logging.Level;
import java.util.logging.Logger;
import objects.Axis;
import objects.Feature;
import objects.FrameBoneRotations;
import objects.FrameJointCoordinates;
import objects.Joint;
import objects.ClassifiableList;
import objects.DistanceMatrix;
import objects.Motion;
import objects.Sample;
import objects.Template;

public abstract class Method implements Serializable {

    private final Axis[] axes = new Axis[]{new Axis("X"), new Axis("Y"), new Axis("Z")};
    private final RetrieverRangeQuery retriever = new RetrieverRangeQuery();

    private List<Double> scoresDBI;
    private List<Double> scoresDI;
    private List<Double> scoresSC;
    private List<Double> scoresFDR;
    private List<Double> scoresCCR;
    private List<Double> scoresEER;
    private List<Double> scoresAUC;
    private List<Double> scoresMAP;
    private List<Double> scoresDCT;
    private List<Double> scoresTD;
    private List<List<Double>> scoresListCMC;
    private List<List<Double>> scoresListFARFRR0;
    private List<List<Double>> scoresListFARFRR1;
    private List<List<Double>> scoresListTARFAR0;
    private List<List<Double>> scoresListTARFAR1;
    private List<List<Double>> scoresListRCLPCN0;
    private List<List<Double>> scoresListRCLPCN1;

    private Classifier classifier;
    private String[] parameterNames;
    private int nFoldsEvaluation;
    private int fineness;

    public Method() {
    }

    public String getName() {
        String[] split = this.getClass().getName().split("\\.");
        return split[split.length - 1].split("Method")[1];
    }

    public Classifier getClassifier() {
        return classifier;
    }

    public Axis[] getAxes() {
        return axes;
    }

    public void setClassifier(Classifier classifier) {
        this.classifier = classifier;
    }

    public void setNFoldsEvaluation(int nFoldsEvaluation) {
        this.nFoldsEvaluation = nFoldsEvaluation;
    }

    public void setFineness(int fineness) {
        this.fineness = fineness;
    }

    // LOAD MOTIONS
    public abstract Motion loadMotion(File fileAMC) throws IOException;

    public List<Motion> loadMotions(File directoryAMC) throws IOException {
        List<Motion> motionsLoaded = new ArrayList();
        for (File fileAMC : directoryAMC.listFiles()) {
            motionsLoaded.add(loadMotion(fileAMC));
        }
        return motionsLoaded;
    }

    // LEARN CLASSIFIER
    public abstract void learnClassifier(List<Motion> motionsLearning);

    public <T extends Sample> Matrix getBetweenClassScatterMatrix(List<T> objects) {
        Matrix meanMatrix = getMeanMatrix(objects);
        int dimension = objects.get(0).getDimension();
        Matrix SigmaB = new Matrix(dimension, dimension);
        List<List<T>> objectsByClass = new ClassifiableList(objects).splitClasses();
        for (List<T> objectsOfClass : objectsByClass) {
            Matrix diff = getMeanMatrix(objectsOfClass).minus(meanMatrix);
            SigmaB.plusEquals(diff.times(diff.transpose()).times((double) objectsOfClass.size() / objects.size()));
        }
        return SigmaB;
    }

    public <T extends Sample> Matrix getWithinClassScatterMatrix(List<T> objects) {
        int dimension = objects.get(0).getDimension();
        Matrix SigmaW = new Matrix(dimension, dimension);
        List<List<T>> objectsByClass = new ClassifiableList(objects).splitClasses();
        for (List<T> objectsOfClass : objectsByClass) {
            Matrix sumOfClass = new Matrix(dimension, dimension);
            for (T object : objectsOfClass) {
                Matrix diff = object.getMatrix().minus(getMeanMatrix(objectsOfClass));
                sumOfClass.plusEquals(diff.times(diff.transpose()));
            }
            SigmaW.plusEquals(sumOfClass.times((double) objectsOfClass.size() / objects.size()));
        }
        return SigmaW;
    }

    public <T extends Sample> Matrix getTotalScatterMatrix(List<T> objects) {
        Matrix meanMatrix = getMeanMatrix(objects);
        int dimension = objects.get(0).getDimension();
        Matrix sumDiff = new Matrix(dimension, dimension);
        for (T object : objects) {
            Matrix diff = object.getMatrix().minus(meanMatrix);
            sumDiff.plusEquals(diff.times(diff.transpose()));
        }
        return sumDiff.times((double) 1 / objects.size());
    }

    public <T extends Sample> Matrix getMeanMatrix(List<T> objects) {
        Matrix meanMatrix = new Matrix(objects.get(0).getDimension(), objects.get(0).getLength());
        for (T object : objects) {
            meanMatrix.plusEquals(object.getMatrix().times((double) 1 / objects.size()));
        }
        return meanMatrix;
    }

    public double[][][] getTensorJointCoordinates(Motion motion) {
        List<Joint> joints = new ArrayList(((FrameJointCoordinates) motion.getFrame(0)).getJoints());
        double[][][] data = new double[motion.getLength()][joints.size()][axes.length];
        for (int l = 0; l < motion.getLength(); l++) {
            for (int j = 0; j < joints.size(); j++) {
                for (int a = 0; a < axes.length; a++) {
                    data[l][j][a] = ((FrameJointCoordinates) motion.getFrame(l)).extractJointAxisCoordinateParameter(axes[a], joints.get(j).getName()).getValue();
                }
            }
        }
        return data;
    }

    public double[][] getTensorBoneRotations(Motion motion) {
        double[][] data = new double[motion.getLength()][((FrameBoneRotations) motion.getFrame(0)).getBoneRotations().length];
        for (int l = 0; l < motion.getLength(); l++) {
            double[] boneRotations = ((FrameBoneRotations) motion.getFrame(l)).getBoneRotations();
            System.arraycopy(boneRotations, 0, data[l], 0, boneRotations.length);
        }
        return data;
    }

    public Matrix getVectorJointCoordinates(Motion motion) {
        double[][][] data3 = getTensorJointCoordinates(motion);
        double[] dataVector = new double[data3.length * data3[0].length * data3[0][0].length];
        int d = 0;
        for (double[][] data2 : data3) {
            for (double[] data1 : data2) {
                for (double data0 : data1) {
                    dataVector[d++] = data0;
                }
            }
        }
        return new Matrix(dataVector, dataVector.length);
    }

    public Matrix getVectorBoneRotations(Motion motion) {
        double[][] data2 = getTensorBoneRotations(motion);
        double[] dataVector = new double[data2.length * data2[0].length];
        int d = 0;
        for (double[] data1 : data2) {
            for (double data0 : data1) {
                dataVector[d++] = data0;
            }
        }
        return new Matrix(dataVector, dataVector.length);
    }

    public void saveClassifier(File fileClassifier) throws IOException {
        ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(fileClassifier));
        oos.writeObject(classifier);
        oos.close();
    }

    public void loadClassifier(File fileClassifier) throws IOException, ClassNotFoundException {
        ObjectInputStream oos = new ObjectInputStream(new FileInputStream(fileClassifier));
        classifier = (Classifier) oos.readObject();
        oos.close();
    }

    // EXTRACT TEMPLATES
    public abstract List<Feature> extractFeatures(Motion motion);

    public List<Template> extractTemplates(List<Motion> motions) {
        List<Template> templates = new ArrayList();
        for (Motion motion : motions) {
            templates.add(extractTemplate(motion));
        }
        return templates;
    }

    public Template extractTemplate(Motion motion) {
        List<Feature> features = extractFeatures(motion);
        int dimension = features.size();
        Matrix matrix = dimension == 0 ? new Matrix(new double[][]{new double[]{}}) : new Matrix(dimension, features.get(0).getLength());
        parameterNames = new String[dimension];
        for (int d = 0; d < dimension; d++) {
            double[] row = features.get(d).getParameterValues();
            matrix.setMatrix(d, d, 0, matrix.getColumnDimension() - 1, new Matrix(row, 1));
            parameterNames[d] = features.get(d).getName();
        }
        return new Template(motion.getSubject(), motion.getId(), matrix);
    }

    //CLASSIFY
    public int classify(Template templateQuery, List<Template> templatesGallery) {
        return classifier.classify(templateQuery, templatesGallery);
    }

    // EVALUATE
    private void initializeEvaluation(int classesEvaluation) {
        scoresDBI = new ArrayList();
        scoresDI = new ArrayList();
        scoresSC = new ArrayList();
        scoresFDR = new ArrayList();
        scoresCCR = new ArrayList();
        scoresEER = new ArrayList();
        scoresAUC = new ArrayList();
        scoresMAP = new ArrayList();
        scoresDCT = new ArrayList();
        scoresTD = new ArrayList();
        scoresListCMC = new ArrayList();
        for (int rank = 0; rank < classesEvaluation; rank++) {
            scoresListCMC.add(new ArrayList());
        }
        scoresListFARFRR0 = new ArrayList();
        scoresListFARFRR1 = new ArrayList();
        scoresListTARFAR0 = new ArrayList();
        scoresListTARFAR1 = new ArrayList();
        scoresListRCLPCN0 = new ArrayList();
        scoresListRCLPCN1 = new ArrayList();
        for (int f = 0; f <= fineness; f++) {
            scoresListFARFRR0.add(new ArrayList());
            scoresListFARFRR1.add(new ArrayList());
            scoresListTARFAR0.add(new ArrayList());
            scoresListTARFAR1.add(new ArrayList());
            scoresListRCLPCN0.add(new ArrayList());
            scoresListRCLPCN1.add(new ArrayList());
        }
    }

    private void printEvaluation() {
        System.out.print("DBI,DI,SC,FDR,CCR,EER,AUC,MAP,DCT,TD\r\n"); // System.out.print("DBI,CCR\r\n"); System.out.print("DBI,DI,SC,FDR,CCR,EER,AUC,MAP,DCT,TD\r\n");
        System.out.print(toString(scoresDBI));
        System.out.print(toString(scoresDI));
        System.out.print(toString(scoresSC));
        System.out.print(toString(scoresFDR));
        System.out.print(toString(scoresCCR));
        System.out.print(toString(scoresEER));
        System.out.print(toString(scoresAUC));
        System.out.print(toString(scoresMAP));
        System.out.print(toString(scoresDCT));
        System.out.print(toString(scoresTD));
        System.out.print("\r\n");
        System.out.print("CMC\r\n");
        for (List<Double> scoresCMC : scoresListCMC) {
            System.out.print(toString(scoresCMC));
            System.out.print("\r\n");
        }
        System.out.print("FAR,FRR,TAR,FAR,RCL,PCN\r\n");
        for (int f = 0; f <= fineness; f++) {
            System.out.print(toString(scoresListFARFRR0.get(f)));
            System.out.print(toString(scoresListFARFRR1.get(f)));
            System.out.print(toString(scoresListTARFAR0.get(f)));
            System.out.print(toString(scoresListTARFAR1.get(f)));
            System.out.print(toString(scoresListRCLPCN0.get(f)));
            System.out.print(toString(scoresListRCLPCN1.get(f)));
            System.out.print("\r\n");
        }
    }

    private String toString(List<Double> scores) {
        return (double) Math.round(getAverage(scores) * 10000) / 10000 + ",";
    }

    public void evaluateHomogeneous(List<Motion> motions, int nClasses, int nFoldsLearning, int attempts) {
        List<List<Motion>> motionsByClass = new ClassifiableList(motions).splitClasses();
        int numberOfClasses = motionsByClass.size();
        if (nClasses < 2) {
            System.out.println("nClasses=" + nClasses + " < 2");
            return;
        }
        if (nClasses > numberOfClasses) {
            System.out.println("nClasses=" + nClasses + " > " + "numberOfClasses=" + numberOfClasses);
            return;
        }
        initializeEvaluation(nClasses);
        Random random = new Random();
        for (int attempt = 0; attempt < attempts; attempt++) {
            List<Integer> indexesClassLearningAndEvaluation = new ArrayList();
            while (indexesClassLearningAndEvaluation.size() < nClasses) {
                int indexClassRandom = random.nextInt(numberOfClasses);
                if (!indexesClassLearningAndEvaluation.contains(indexClassRandom)) {
                    indexesClassLearningAndEvaluation.add(indexClassRandom);
                }
            }
            List<Motion> motionsLearningAndEvaluation = new ArrayList();
            for (int indexClassLearningAndEvaluation : indexesClassLearningAndEvaluation) {
                motionsLearningAndEvaluation.addAll(motionsByClass.get(indexClassLearningAndEvaluation));
            }
            List<List<Motion>> foldsLearning = new ArrayList();
            for (int fFoldLearning = 0; fFoldLearning < nFoldsLearning; fFoldLearning++) {
                foldsLearning.add(new ArrayList());
            }
            for (int i = 0; i < motionsLearningAndEvaluation.size(); i++) {
                foldsLearning.get(i % nFoldsLearning).add(motionsLearningAndEvaluation.get(i));
            }
            for (int fFoldLearning = 0; fFoldLearning < nFoldsLearning; fFoldLearning++) {
                List<Motion> motionsLearning = new ArrayList();
                List<Motion> motionsEvaluation = new ArrayList();
                for (Motion motion : foldsLearning.get(fFoldLearning)) {
                    motionsLearning.add(motion);
                }
                for (int f = 0; f < nFoldsLearning; f++) {
                    if (f != fFoldLearning) {
                        for (Motion motion : foldsLearning.get(f)) {
                            motionsEvaluation.add(motion);
                        }
                    }
                }
//                System.out.println("attempt=" + attempt + " fFoldLearning=" + fFoldLearning);
                evaluate(motionsLearning, motionsEvaluation);
            }
        }
        printEvaluation();
    }

    public void evaluateHeterogeneous(List<Motion> motions, int nClassesLearning, int nClassesEvaluation, int attempts) {
        List<List<Motion>> motionsByClass = new ClassifiableList(motions).splitClasses();
        int numberOfClasses = motionsByClass.size();
        if (nClassesLearning < 2) {
            System.out.println("nClassesLearning=" + nClassesLearning + " < 2");
            return;
        }
        if (nClassesEvaluation < 2) {
            System.out.println("nClassesEvaluation=" + nClassesEvaluation + " < 2");
            return;
        }
        if (nClassesLearning + nClassesEvaluation > numberOfClasses) {
            System.out.println("nClassesLearning=" + nClassesLearning + " + " + "nClassesEvaluation=" + nClassesEvaluation + " > " + "numberOfClasses=" + numberOfClasses);
            return;
        }
        initializeEvaluation(nClassesEvaluation);
        Random random = new Random();
        for (int attempt = 0; attempt < attempts; attempt++) {
            List<Integer> indexesClassLearning = new ArrayList();
            List<Integer> indexesClassEvaluation = new ArrayList();
            while (indexesClassLearning.size() < nClassesLearning) {
                int indexClassRandom = random.nextInt(numberOfClasses);
                if (!indexesClassLearning.contains(indexClassRandom)) {
                    indexesClassLearning.add(indexClassRandom);
                }
            }
            while (indexesClassEvaluation.size() < nClassesEvaluation) {
                int indexClassRandom = random.nextInt(numberOfClasses);
                if (!indexesClassEvaluation.contains(indexClassRandom) && !indexesClassLearning.contains(indexClassRandom)) {
                    indexesClassEvaluation.add(indexClassRandom);
                }
            }
            List<Motion> motionsLearning = new ArrayList();
            List<Motion> motionsEvaluation = new ArrayList();
            for (int indexClassLearning : indexesClassLearning) {
                for (Motion motion : motionsByClass.get(indexClassLearning)) {
                    motionsLearning.add(motion);
                }
            }
            for (int indexClassEvaluation : indexesClassEvaluation) {
                for (Motion motion : motionsByClass.get(indexClassEvaluation)) {
                    motionsEvaluation.add(motion);
                }
            }
//                System.out.println("attempt=" + attempt + " indexesClassLearning=" + indexesClassLearning + " indexesClassEvaluation=" + indexesClassEvaluation);
            evaluate(motionsLearning, motionsEvaluation);
        }
        printEvaluation();
    }

    private void evaluate(List<Motion> motionsLearning, List<Motion> motionsEvaluation) {
        learnClassifier(motionsLearning);
        List<Template> templatesEvaluation = extractTemplates(motionsEvaluation);
        int numberOfClassesEvaluation = new ClassifiableList(templatesEvaluation).splitClasses().size();
        List<Double> scoresAttemptCCR = new ArrayList();
        List<Double> scoresAttemptEER = new ArrayList();
        List<Double> scoresAttemptAUC = new ArrayList();
        List<Double> scoresAttemptMAP = new ArrayList();
        List<Double> scoresAttemptDCT = new ArrayList();
        List<Double> scoresAttemptTD = new ArrayList();
        List<List<Double>> scoresListAttemptCMC = new ArrayList();
        for (int rank = 0; rank < numberOfClassesEvaluation; rank++) {
            scoresListAttemptCMC.add(new ArrayList());
        }
        List<List<Double>> scoresListAttemptFARFRR0 = new ArrayList();
        List<List<Double>> scoresListAttemptFARFRR1 = new ArrayList();
        List<List<Double>> scoresListAttemptTARFAR0 = new ArrayList();
        List<List<Double>> scoresListAttemptTARFAR1 = new ArrayList();
        List<List<Double>> scoresListAttemptRCLPCN0 = new ArrayList();
        List<List<Double>> scoresListAttemptRCLPCN1 = new ArrayList();
        for (int f = 0; f <= fineness; f++) {
            scoresListAttemptFARFRR0.add(new ArrayList());
            scoresListAttemptFARFRR1.add(new ArrayList());
            scoresListAttemptTARFAR0.add(new ArrayList());
            scoresListAttemptTARFAR1.add(new ArrayList());
            scoresListAttemptRCLPCN0.add(new ArrayList());
            scoresListAttemptRCLPCN1.add(new ArrayList());
        }
        long beginning = System.currentTimeMillis();
        setDistanceMatrix(templatesEvaluation);
        scoresAttemptDCT.add((System.currentTimeMillis() - beginning) / Math.pow(templatesEvaluation.size(), 2));
        scoresAttemptTD.add(getTemplateDimensionality(templatesEvaluation));
        List<List<Template>> foldsEvaluation = new ArrayList();
        for (int fFoldEvaluation = 0; fFoldEvaluation < nFoldsEvaluation; fFoldEvaluation++) {
            foldsEvaluation.add(new ArrayList());
        }
        for (int i = 0; i < templatesEvaluation.size(); i++) {
            foldsEvaluation.get(i % nFoldsEvaluation).add(templatesEvaluation.get(i));
        }
        for (int fFoldEvaluation = 0; fFoldEvaluation < nFoldsEvaluation; fFoldEvaluation++) {
            List<Template> templatesTest = foldsEvaluation.get(fFoldEvaluation);
            List<Template> templatesGallery = new ArrayList();
            for (int f = 0; f < nFoldsEvaluation; f++) {
                if (f != fFoldEvaluation) {
                    templatesGallery.addAll(foldsEvaluation.get(f));
                }
            }
            double[] cumulativeMatchCharacteristic = getCumulativeMatchCharacteristic(templatesTest, templatesGallery);
            for (int rank = 0; rank < numberOfClassesEvaluation; rank++) {
                scoresListAttemptCMC.get(rank).add(cumulativeMatchCharacteristic[rank]);
            }
            scoresAttemptCCR.add(cumulativeMatchCharacteristic[0]);
            double highRadiusFARFRR = 1;
            while (getFARFRR(highRadiusFARFRR, templatesTest, templatesGallery)[1] > 0) {
//                System.out.print("findHighRadiusFARFRR: highRadiusFARFRR=" + highRadiusFARFRR + "\r\n");
                highRadiusFARFRR *= 2;
            }
            double radiusEER = findRadiusEER(0, highRadiusFARFRR, templatesTest, templatesGallery);
            scoresAttemptEER.add(getFARFRR(radiusEER, templatesTest, templatesGallery)[0]);
            for (int f = 0; f <= fineness; f++) {
                double radiusFARFRR = f * 2 * radiusEER / fineness;
                double[] ratesFARFRR = getFARFRR(radiusFARFRR, templatesTest, templatesGallery);
                scoresListAttemptFARFRR0.get(f).add(ratesFARFRR[0]);
                scoresListAttemptFARFRR1.get(f).add(ratesFARFRR[1]);
//                System.out.print("FAR/FRR: " + (100 * f / (fineness + 1)) + "% radiusFARFRR=" + radiusFARFRR + "\r\n");
            }
            double highRadiusTARFAR = 1;
            while (getTARFAR(highRadiusTARFAR, templatesTest, templatesGallery)[0] < 1 || getTARFAR(highRadiusTARFAR, templatesTest, templatesGallery)[1] < 1) {
//                System.out.print("findHighRadiusGARIAR: highRadiusTARFAR=" + highRadiusTARFAR + "\r\n");
                highRadiusTARFAR *= 2;
            }
            double maxRadiusTARFAR = findMaxRadiusGARIAR(0, highRadiusTARFAR, templatesTest, templatesGallery);
            List<double[]> curveTARFAR = new ArrayList();
            List<double[]> gapsFAR = new ArrayList();
            double[] minRatesTARFAR = getTARFAR(0, templatesTest, templatesGallery);
            double[] maxRatesTARFAR = getTARFAR(maxRadiusTARFAR, templatesTest, templatesGallery);
            curveTARFAR.add(minRatesTARFAR);
            curveTARFAR.add(maxRatesTARFAR);
            gapsFAR.add(new double[]{minRatesTARFAR[1], maxRatesTARFAR[1], 0, maxRadiusTARFAR});
            for (int f = 1; f < fineness; f++) {
                double maxGapSize = 0;
                double radiusLow = 0;
                double radiusHigh = 0;
                double radiusTARFAR = 0;
                int maxGapIndex = 0;
                for (int gapIndex = 0; gapIndex < gapsFAR.size(); gapIndex++) {
                    double[] gap = gapsFAR.get(gapIndex);
                    double gapSize = gap[1] - gap[0];
                    if (maxGapSize < gapSize) {
                        maxGapSize = gapSize;
                        radiusLow = gap[2];
                        radiusHigh = gap[3];
                        radiusTARFAR = (radiusLow + radiusHigh) / 2;
                        maxGapIndex = gapIndex;
                    }
                }
                double[] maxGap = gapsFAR.get(maxGapIndex);
                double[] ratesTARFAR = getTARFAR(radiusTARFAR, templatesTest, templatesGallery);
                gapsFAR.remove(maxGap);
                gapsFAR.add(new double[]{maxGap[0], ratesTARFAR[1], radiusLow, radiusTARFAR});
                gapsFAR.add(new double[]{ratesTARFAR[1], maxGap[1], radiusTARFAR, radiusHigh});
                curveTARFAR.add(ratesTARFAR);
//                System.out.print("TAR/FAR: " + (100 * (f + 2) / (fineness + 1)) + "% radiusTARFAR=" + radiusTARFAR + "\r\n");
            }
            curveTARFAR = sortCurve(curveTARFAR, 1);
            for (int f = 0; f <= fineness; f++) {
                scoresListAttemptTARFAR0.get(f).add(curveTARFAR.get(f)[0]);
                scoresListAttemptTARFAR1.get(f).add(curveTARFAR.get(f)[1]);
            }
            scoresAttemptAUC.add(getAreaUnderCurve(curveTARFAR, 1));
            double highRadiusRCLPCN = 1;
            while (getRCLPCN(highRadiusRCLPCN, templatesTest, templatesGallery)[0] < 1) {
//                System.out.print("findHighRadiusRCLPCN: highRadiusRCLPCN=" + highRadiusRCLPCN + "\r\n");
                highRadiusRCLPCN *= 2;
            }
            double maxRadiusRCLPCN = findMaxRadiusRCLPCN(0, highRadiusRCLPCN, templatesTest, templatesGallery);
            List<double[]> curveRCLPCN = new ArrayList();
            List<double[]> gapsRCL = new ArrayList();
            double[] minRatesRCLPCN = getRCLPCN(0, templatesTest, templatesGallery);
            double[] maxRatesRCLPCN = getRCLPCN(maxRadiusRCLPCN, templatesTest, templatesGallery);
            curveRCLPCN.add(minRatesRCLPCN);
            curveRCLPCN.add(maxRatesRCLPCN);
            gapsRCL.add(new double[]{minRatesRCLPCN[0], maxRatesRCLPCN[0], 0, maxRadiusRCLPCN});
            for (int f = 1; f < fineness; f++) {
                double maxGapSize = 0;
                double radiusLow = 0;
                double radiusHigh = 0;
                double radiusRCLPCN = 0;
                int maxGapIndex = 0;
                for (int gapIndex = 0; gapIndex < gapsRCL.size(); gapIndex++) {
                    double[] gap = gapsRCL.get(gapIndex);
                    double gapSize = gap[1] - gap[0];
                    if (maxGapSize < gapSize) {
                        maxGapSize = gapSize;
                        radiusLow = gap[2];
                        radiusHigh = gap[3];
                        radiusRCLPCN = (radiusLow + radiusHigh) / 2;
                        maxGapIndex = gapIndex;
                    }
                }
                double[] maxGap = gapsRCL.get(maxGapIndex);
                double[] ratesRCLPCN = getRCLPCN(radiusRCLPCN, templatesTest, templatesGallery);
                gapsRCL.remove(maxGap);
                gapsRCL.add(new double[]{maxGap[0], ratesRCLPCN[0], radiusLow, radiusRCLPCN});
                gapsRCL.add(new double[]{ratesRCLPCN[0], maxGap[1], radiusRCLPCN, radiusHigh});
                curveRCLPCN.add(ratesRCLPCN);
//                System.out.print("RCL/PCN: " + (100 * (f + 2) / (fineness + 1)) + "% radiusRCLPCN=" + radiusRCLPCN + "\indexClassRandom\n");
            }
            curveRCLPCN = sortCurve(curveRCLPCN, 0);
            for (int f = 0; f <= fineness; f++) {
                scoresListAttemptRCLPCN0.get(f).add(curveRCLPCN.get(f)[0]);
                scoresListAttemptRCLPCN1.get(f).add(curveRCLPCN.get(f)[1]);
            }
            scoresAttemptMAP.add(getAreaUnderCurve(curveRCLPCN, 0));
        }
        scoresDBI.add(getDaviesBouldinIndex(templatesEvaluation));
        scoresDI.add(getDunnIndex(templatesEvaluation));
        scoresSC.add(getSilhouetteCoefficient(templatesEvaluation));
        scoresFDR.add(getFishersDiscriminantRatio(templatesEvaluation));
        scoresCCR.add(getAverage(scoresAttemptCCR));
        scoresEER.add(getAverage(scoresAttemptEER));
        scoresAUC.add(getAverage(scoresAttemptAUC));
        scoresMAP.add(getAverage(scoresAttemptMAP));
        scoresDCT.add(getAverage(scoresAttemptDCT));
        scoresTD.add(getAverage(scoresAttemptTD));
        for (int rank = 0; rank < numberOfClassesEvaluation; rank++) {
            scoresListCMC.get(rank).add(getAverage(scoresListAttemptCMC.get(rank)));
        }
        for (int f = 0; f <= fineness; f++) {
            scoresListFARFRR0.get(f).add(getAverage(scoresListAttemptFARFRR0.get(f)));
            scoresListFARFRR1.get(f).add(getAverage(scoresListAttemptFARFRR1.get(f)));
            scoresListTARFAR0.get(f).add(getAverage(scoresListAttemptTARFAR0.get(f)));
            scoresListTARFAR1.get(f).add(getAverage(scoresListAttemptTARFAR1.get(f)));
            scoresListRCLPCN0.get(f).add(getAverage(scoresListAttemptRCLPCN0.get(f)));
            scoresListRCLPCN1.get(f).add(getAverage(scoresListAttemptRCLPCN1.get(f)));
        }
    }

    private List<double[]> sortCurve(List<double[]> curve, int index) {
        List<double[]> sortedCurve = new ArrayList();
        int length = curve.size();
        for (int l = 0; l < length; l++) {
            double minPointValue = Double.MAX_VALUE;
            int minPointIndex = 0;
            for (int pointIndex = 0; pointIndex < curve.size(); pointIndex++) {
                if (minPointValue > curve.get(pointIndex)[index]) {
                    minPointValue = curve.get(pointIndex)[index];
                    minPointIndex = pointIndex;
                }
            }
            sortedCurve.add(curve.get(minPointIndex));
            curve.remove(minPointIndex);
        }
        return sortedCurve;
    }

    private double getAverage(List<Double> scores) {
        double sum = 0;
        for (double score : scores) {
            sum += score;
        }
        return sum / scores.size();
    }

    private void setDistanceMatrix(List<Template> templatesEvaluation) {
        DistanceMatrix distanceMatrix = new DistanceMatrix(classifier.getDistanceTemplates());
        int size = templatesEvaluation.size();
        long beginning = System.currentTimeMillis();//
        int k = 0;//
        for (int i = 0; i < size; i++) {
            for (int j = 0; j < size; j++) {
                distanceMatrix.putDistance(templatesEvaluation.get(i), templatesEvaluation.get(j));
                int pBefore = (100 * k) / (size * size);//
                k++;//
                int pAfter = (100 * k) / (size * size);//
                long time = System.currentTimeMillis() - beginning;//
                if (pBefore < pAfter) {//
                    System.out.println(Executor.printTime(time) + ": " + pAfter + "%, remaining: " + Executor.printTime(100 * time / pAfter - time));//
                }//
            }
        }
        classifier.setDistanceMatrix(distanceMatrix);
        retriever.setDistanceMatrix(distanceMatrix);
    }

    private double[] getCumulativeMatchCharacteristic(List<Template> templatesTest, List<Template> templatesGallery) {
        List<List<Template>> split = new ClassifiableList(templatesGallery).splitClasses();
        int classes = split.size();
        int number = templatesTest.size();
        double[] cmc = new double[classes];
        for (Template templateTest : templatesTest) {
            List<Template> templatesGalleryRank = new ArrayList();
            templatesGalleryRank.addAll(templatesGallery);
            int[] ranking = new int[classes];
            for (int rank = 0; rank < classes; rank++) {
                int classifiedSubject = classifier.classify(templateTest, templatesGalleryRank);
                ranking[rank] = classifiedSubject;
                int i = 0;
                while (split.get(i).get(0).getSubject() != classifiedSubject) {
                    i++;
                }
                templatesGalleryRank.removeAll(split.get(i));
            }
            for (int rank = 0; rank < classes; rank++) {
                boolean isIn = false;
                for (int r = 0; r <= rank; r++) {
                    if (ranking[r] == templateTest.getSubject()) {
                        isIn = true;
                    }
                }
                if (isIn) {
                    cmc[rank] += (double) 1 / number;
                }
            }
        }
        return cmc;
    }

    private double getDaviesBouldinIndex(List<Template> templatesEvaluation) { // low is good
        List<List<Template>> split = new ClassifiableList(templatesEvaluation).splitClasses();
        int classes = split.size();
        double sum = 0;
        for (int c1 = 0; c1 < classes; c1++) {
            List<Template> classOfTemplates1 = split.get(c1);
            double maxFrac = 0;
            for (int c2 = c1 + 1; c2 < classes; c2++) {
                List<Template> classOfTemplates2 = split.get(c2);
                double frac = (getAverageDistanceFromCentroid(classOfTemplates1) + getAverageDistanceFromCentroid(classOfTemplates2)) / getDistanceBetweenCentroids(classOfTemplates1, classOfTemplates2);
                if (maxFrac < frac) {
                    maxFrac = frac;
                }
            }
            sum += maxFrac;
        }
        return sum;
    }

    private double getDunnIndex(List<Template> templatesEvaluation) { // high is good
        List<List<Template>> split = new ClassifiableList(templatesEvaluation).splitClasses();
        int classes = split.size();
        double min = Double.MAX_VALUE;
        for (int c1 = 0; c1 < classes; c1++) {
            for (int c2 = c1 + 1; c2 < classes; c2++) {
                double dist = getDistanceBetweenCentroids(split.get(c1), split.get(c2));
                if (min > dist) {
                    min = dist;
                }
            }
        }
        double max = Double.MAX_VALUE;
        for (int c = 0; c < classes; c++) {
            double dist = getAverageDistanceFromCentroid(split.get(c));
            if (max > dist) {
                max = dist;
            }
        }
        return min / max;
    }

    private double getSilhouetteCoefficient(List<Template> templatesEvaluation) { // high is good
        List<List<Template>> split = new ClassifiableList(templatesEvaluation).splitClasses();
        DistanceMatrix distanceMatrix = classifier.getDistanceMatrix();
        double sum = 0;
        for (List<Template> thisClass : split) {
            for (Template thisTemplate : thisClass) {
                double sumA = 0;
                for (Template otherTemplateOfThisClass : thisClass) {
                    double dist = distanceMatrix.getDistance(thisTemplate, otherTemplateOfThisClass);
                    sumA += dist;
                }
                double avgA = sumA / thisClass.size();
                double minAvgB = Double.MAX_VALUE;
                for (List<Template> otherClass : split) {
                    if (!thisClass.equals(otherClass)) {
                        double sumB = 0;
                        for (Template otherTemplateOfOtherClass : otherClass) {
                            double dist = distanceMatrix.getDistance(thisTemplate, otherTemplateOfOtherClass);
                            sumB += dist;
                        }
                        double avgB = sumB / otherClass.size();
                        if (minAvgB > avgB) {
                            minAvgB = avgB;
                        }
                    }
                }
                sum += (minAvgB - avgA) / Math.max(avgA, minAvgB);
            }
        }
        return sum / templatesEvaluation.size();
    }

    private double getFishersDiscriminantRatio(List<Template> templatesEvaluation) { // high is good
        List<List<Template>> split = new ClassifiableList(templatesEvaluation).splitClasses();
        return getBetweenClassScatter(split) / getWithinClassScatter(split);
    }

    private double getBetweenClassScatter(List<List<Template>> templatesByClass) {
        DistanceMatrix distanceMatrix = classifier.getDistanceMatrix();
        double sum = 0;
        List<Template> centroids = new ArrayList();
        for (List<Template> templatesOfClass : templatesByClass) {
            centroids.add(getCentroid(templatesOfClass));
        }
        Template centroidOverall = getCentroid(centroids);
        for (Template centroid : centroids) {
            sum += distanceMatrix.getDistance(centroidOverall, centroid);
        }
        return sum / templatesByClass.size();
    }

    private double getWithinClassScatter(List<List<Template>> templatesByClass) {
        DistanceMatrix distanceMatrix = classifier.getDistanceMatrix();
        double sum = 0;
        int num = 0;
        for (List<Template> templatesOfClass : templatesByClass) {
            Template centroid = getCentroid(templatesOfClass);
            num += templatesOfClass.size();
            for (Template template : templatesOfClass) {
                sum += distanceMatrix.getDistance(centroid, template);
            }
        }
        return sum / num;
    }

    private double getDistanceBetweenCentroids(List<Template> templates1, List<Template> templates2) {
        return classifier.getDistanceMatrix().getDistance(getCentroid(templates1), getCentroid(templates2));
    }

    private double getAverageDistanceFromCentroid(List<Template> templates) {
        DistanceMatrix distanceMatrix = classifier.getDistanceMatrix();
        Template centroid = getCentroid(templates);
        double sum = 0;
        for (Template template : templates) {
            sum += distanceMatrix.getDistance(centroid, template);
        }
        return sum / templates.size();
    }

    private Template getCentroid(List<Template> templates) {
        DistanceMatrix distanceMatrix = classifier.getDistanceMatrix();
        Template centroid = new Template(0, "", null);
        if (templates.isEmpty()) {
            System.out.println("empty class");
        } else {
            double minSum = Double.MAX_VALUE;
            for (Template centroidCandidate : templates) {
                double sum = 0;
                for (Template template : templates) {
                    sum += distanceMatrix.getDistance(centroidCandidate, template);
                }
                if (minSum > sum) {
                    minSum = sum;
                    centroid = centroidCandidate;
                }
            }
        }
        return centroid;
    }

    public double getTemplateDimensionality(List<Template> templatesEvaluation) {
        int sum = 0;
        for (Template template : templatesEvaluation) {
            sum += template.getDimension() * template.getLength();
        }
        return (double) sum / templatesEvaluation.size();
    }

    public double[] getFARFRR(double radius, List<Template> templatesTest, List<Template> templatesGallery) {
        double far = 0;
        double frr = 0;
        int countGenuine = 0;
        int countImpostor = 0;
        for (Template templateTest : templatesTest) {
            List<Template> templatesGenuine = new ArrayList();
            List<Template> templatesImpostor = new ArrayList();
            for (Template templateGallery : templatesGallery) {
                if (templateGallery.getSubject() == templateTest.getSubject()) {
                    templatesGenuine.add(templateGallery);
                } else {
                    templatesImpostor.add(templateGallery);
                }
            }
            retriever.setR(radius);
            List<Template> templatesRetrieved = retriever.retrieve(templateTest, templatesGallery);
            for (Template templateRetrieved : templatesRetrieved) {
                if (!templatesGenuine.contains(templateRetrieved)) {
                    far++;
                }
            }
            for (Template templateGenuine : templatesGenuine) {
                if (!templatesRetrieved.contains(templateGenuine)) {
                    frr++;
                }
            }
            countGenuine += templatesGenuine.size();
            countImpostor += templatesImpostor.size();
        }
        far /= countImpostor;
        frr /= countGenuine;
        return new double[]{far, frr};
    }

    private double findRadiusEER(double beg, double end, List<Template> templatesTest, List<Template> templatesGallery) {
        double radiusRCLPCN = (end + beg) / 2;
//        System.out.print("findRadiusEER: radiusRCLPCN=" + radiusRCLPCN + "\r\n");
        double[] ratesFARFRR = getFARFRR(radiusRCLPCN, templatesTest, templatesGallery);
        if (end - beg > 1) {
            if (ratesFARFRR[0] > ratesFARFRR[1]) {
                return findRadiusEER(beg, radiusRCLPCN, templatesTest, templatesGallery);
            } else {
                return findRadiusEER(radiusRCLPCN, end, templatesTest, templatesGallery);
            }
        }
        return radiusRCLPCN;
    }

    public double[] getTARFAR(double radius, List<Template> templatesTest, List<Template> templatesGallery) {
        double gar = 0;
        double iar = 0;
        int countGenuine = 0;
        int countImpostor = 0;
        for (Template templateTest : templatesTest) {
            List<Template> templatesGenuine = new ArrayList();
            List<Template> templatesImpostor = new ArrayList();
            for (Template templateGallery : templatesGallery) {
                if (templateGallery.getSubject() == templateTest.getSubject()) {
                    templatesGenuine.add(templateGallery);
                } else {
                    templatesImpostor.add(templateGallery);
                }
            }
            retriever.setR(radius);
            List<Template> templatesRetrieved = retriever.retrieve(templateTest, templatesGallery);
            for (Template templateRetrieved : templatesRetrieved) {
                if (templatesGenuine.contains(templateRetrieved)) {
                    gar++;
                }
                if (templatesImpostor.contains(templateRetrieved)) {
                    iar++;
                }
            }
            countGenuine += templatesGenuine.size();
            countImpostor += templatesImpostor.size();
        }
        iar /= countImpostor;
        gar /= countGenuine;
        return new double[]{gar, iar};
    }

    private double findMaxRadiusGARIAR(double beg, double end, List<Template> templatesTest, List<Template> templatesGallery) {
        double radiusGARIAR = (end + beg) / 2;
//        System.out.print("findMaxRadiusGARIAR: highRadiusTARFAR=" + highRadiusTARFAR + "\r\n");
        double[] ratesGARIAR = getTARFAR(radiusGARIAR, templatesTest, templatesGallery);
        if (end - beg > 1) {
            if (ratesGARIAR[0] == 1 && ratesGARIAR[1] == 1) {
                return findMaxRadiusGARIAR(beg, radiusGARIAR, templatesTest, templatesGallery);
            } else {
                return findMaxRadiusGARIAR(radiusGARIAR, end, templatesTest, templatesGallery);
            }
        }
        return end;
    }

    private double getAreaUnderCurve(List<double[]> curve, int index) {
        double auc = 0;
        for (int l = 1; l < curve.size(); l++) {
            auc += 0.5 * Math.abs(curve.get(l)[index] - curve.get(l - 1)[index]) * Math.abs(curve.get(l)[1 - index] + curve.get(l - 1)[1 - index]);
        }
        return auc;
    }

    public double[] getRCLPCN(double radius, List<Template> templatesTest, List<Template> templatesGallery) {
        double rcl = 0;
        double pcn = 0;
        int countRetrieved = 0;
        int countGenuine = 0;
        for (Template templateTest : templatesTest) {
            List<Template> templatesGenuine = new ArrayList();
            for (Template templateGallery : templatesGallery) {
                if (templateGallery.getSubject() == templateTest.getSubject()) {
                    templatesGenuine.add(templateGallery);
                }
            }
            retriever.setR(radius);
            List<Template> templatesRetrieved = retriever.retrieve(templateTest, templatesGallery);
            for (Template templateRetrieved : templatesRetrieved) {
                if (templatesGenuine.contains(templateRetrieved)) {
                    pcn++;
                }
            }
            for (Template templateGenuine : templatesGenuine) {
                if (templatesRetrieved.contains(templateGenuine)) {
                    rcl++;
                }
            }
            countRetrieved += templatesRetrieved.size();
            countGenuine += templatesGenuine.size();
        }
        pcn = countRetrieved == 0 ? 1 : pcn / countRetrieved;
        rcl /= countGenuine;
        return new double[]{rcl, pcn};
    }

    private double findMaxRadiusRCLPCN(double beg, double end, List<Template> templatesTest, List<Template> templatesGallery) {
        double radiusRCLPCN = (end + beg) / 2;
//        System.out.print("findMaxRadiusRCLPCN: radiusRCLPCN=" + radiusRCLPCN + "\r\n");
        double[] ratesRCLPCN = getRCLPCN(radiusRCLPCN, templatesTest, templatesGallery);
        if (end - beg > 1) {
            if (ratesRCLPCN[0] == 1) {
                return findMaxRadiusRCLPCN(beg, radiusRCLPCN, templatesTest, templatesGallery);
            } else {
                return findMaxRadiusRCLPCN(radiusRCLPCN, end, templatesTest, templatesGallery);
            }
        }
        return end;
    }

    public double distanceL2Squared(double[] v1, double[] v2) {
        if (v1.length > v2.length) {
            return distanceL2Squared(v2, v1);
        }
        double dist = 0;
        for (int d = 0; d < v1.length; d++) {
            dist += Math.pow(v1[d] - v2[d], 2);
        }
        return dist;
    }

    public double pointToLineDistance(double[] p, double[] p0, double[] p1) {
        double[] v = new double[p0.length];
        double[] w = new double[p0.length];
        for (int i = 0; i < p0.length; i++) {
            v[i] = p1[i] - p0[i];
            w[i] = p[i] - p0[i];
        }
        double b = dotProduct(w, v) / dotProduct(v, v);
        double dist = 0;
        for (int i = 0; i < v.length; i++) {
            dist += (w[i] - b * v[i]) * (w[i] - b * v[i]);
        }
        return Math.sqrt(dist);
    }

    public double dotProduct(double[] v1, double[] v2) {
        double rtv = 0;
        for (int i = 0; i < v1.length; i++) {
            rtv += v1[i] * v2[i];
        }
        return rtv;
    }

    private void exportTemplatesToCSV(String fileNameCSV, List<Template> templates) {
        try {
            System.out.print("saving " + fileNameCSV + "\r\n");
            FileWriter writer = new FileWriter(fileNameCSV);
            Template template0 = templates.get(0);
            writer.append("subject").append(",");
            for (int l = 0; l < template0.getLength(); l++) {
                for (String parameterName : parameterNames) {
                    writer.append(String.valueOf(l)).append(":").append(parameterName).append(",");
                }
            }
            writer.append("\r\n");
            for (Template template : templates) {
                writer.append(String.valueOf(template.getSubject()));
                for (int l = 0; l < template.getLength(); l++) {
                    for (int d = 0; d < template.getDimension(); d++) {
                        writer.append(",").append(String.valueOf(template.getParameterValue(d, l)));
                    }
                }
                writer.append("\r\n");
            }
            writer.flush();
            writer.close();
        } catch (IOException ex) {
            Logger.getLogger(Method.class.getName()).log(Level.SEVERE, null, ex);
        }
    }
}
