package methods;

import Jama.Matrix;
import Jama.SingularValueDecomposition;
import algorithms.Classifier;
import algorithms.Classifier1NN;
import algorithms.ClassifierLinear;
import algorithms.DistanceTemplatesBaseline;
import algorithms.DistanceVectorsMahalanobis;
import executor.Executor;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import objects.ClassifiableList;
import objects.Feature;
import objects.Motion;
import objects.Parameter;
import objects.Sample;

public class Method_MMC extends Method {

    private final boolean dataType; // true = BoneRotations, false = JointCoordinates
    private final Classifier classifier = new Classifier1NN();
    private final int length = 150;

    public Method_MMC(boolean dataType) {
        this.dataType = dataType;
    }

    @Override
    public Motion loadMotion(File fileAMC) throws IOException {
        return dataType ? Executor.loadMotionBoneRotations(fileAMC) : Executor.loadMotionJointCoordinates(fileAMC);
    }

    private Matrix getVector(Motion motion) {
        return dataType ? getVectorBoneRotations(motion) : getVectorJointCoordinates(motion);
    }

    @Override
    public void learnClassifier(List<Motion> motionsLearning) {

        // TRANSFORMATION MATRIX
        List<Sample> samples = new ArrayList();
        for (Motion motion : motionsLearning) {
            motion.adjust(length);
            samples.add(new Sample(motion.getSubject(), motion.getId(), getVector(motion)));
        }
        int numberOfSamples = samples.size();
        int dimensionSample = samples.get(0).getDimension();
        List<List<Sample>> samplesByClass = new ClassifiableList(samples).splitClasses();
        int numberOfClasses = samplesByClass.size();
        Matrix meanSampleMatrix = getMeanMatrix(samples);
        Matrix Chi = new Matrix(dimensionSample, numberOfSamples);
        for (int n = 0; n < numberOfSamples; n++) {
            double[] column = (samples.get(n).getMatrix().minus(meanSampleMatrix)).getColumnPackedCopy();
            Chi.setMatrix(0, dimensionSample - 1, n, n, new Matrix(column, column.length));
        }
        Chi.timesEquals((double) 1 / Math.sqrt(numberOfSamples));
        Matrix Upsilon = new Matrix(dimensionSample, numberOfClasses);
        for (int c = 0; c < numberOfClasses; c++) {
            List<Sample> samplesOfClass = samplesByClass.get(c);
            double[] column = (getMeanMatrix(samplesOfClass).minus(meanSampleMatrix)).times(Math.sqrt((double) samplesOfClass.size() / numberOfSamples)).getColumnPackedCopy();
            Upsilon.setMatrix(0, dimensionSample - 1, c, c, new Matrix(column, column.length));
        }
        SingularValueDecomposition svdChi = new SingularValueDecomposition(Chi);
        Matrix Omega = svdChi.getU();
        Matrix ThetaInverseSquareRoot = svdChi.getS();
        int dim = Math.min(ThetaInverseSquareRoot.getRowDimension(), ThetaInverseSquareRoot.getColumnDimension());
        for (int d = 0; d < dim; d++) {
            ThetaInverseSquareRoot.set(d, d, (double) 1 / Math.sqrt(ThetaInverseSquareRoot.get(d, d)));
        }
        SingularValueDecomposition svdThetaOmegaUpsilon = new SingularValueDecomposition(ThetaInverseSquareRoot.times(Omega.transpose()).times(Upsilon));
        Matrix Xi = svdThetaOmegaUpsilon.getU();
        Matrix Psi = Omega.times(ThetaInverseSquareRoot).times(Xi);
        Matrix Delta = Psi.transpose().times(getBetweenClassScatterMatrix(samples)).times(Psi);
        List<Integer> indexes = new ArrayList();
        for (int c = 0; c < numberOfClasses; c++) {
            if (Delta.get(c, c) >= 0.5) {
                indexes.add(c);
            }
        }
        int dimensionTemplate = indexes.size();
        Matrix Phi = new Matrix(Psi.getRowDimension(), dimensionTemplate);
        for (int d = 0; d < dimensionTemplate; d++) {
            int index = indexes.get(d);
            double[] column = Psi.getMatrix(0, dimensionSample - 1, index, index).getColumnPackedCopy();
            Phi.setMatrix(0, dimensionSample - 1, d, d, new Matrix(column, column.length));
        }
        classifier.setTransformationMatrix(Phi);
        setClassifier(classifier);

        // DISTANCE FUNCTION
        DistanceTemplatesBaseline distanceTemplates = new DistanceTemplatesBaseline();
        DistanceVectorsMahalanobis distanceVectors = new DistanceVectorsMahalanobis();
        distanceVectors.setMatrix(getTotalScatterMatrix(extractTemplates(motionsLearning)).inverse());
        distanceTemplates.setDistanceVectors(distanceVectors);
        classifier.setDistanceTemplates(distanceTemplates);
        setClassifier(classifier);
    }

    @Override
    public List<Feature> extractFeatures(Motion motion) {
        motion.adjust(length);
        double[][] data = getClassifier().getTransformationMatrix().transpose().times(getVector(motion)).getArray();
        List<Feature> features = new ArrayList();
        for (int d = 0; d < data.length; d++) {
            features.add(new Feature(new Parameter[]{new Parameter(String.valueOf(d), data[d][0])}));
        }
        return features;
    }
}
